@@ -1,6 +1,6 @@
 //! Framed transport dispatcher
 use std::task::{Context, Poll, Waker};
-use std::{cell::Cell, cell::RefCell, future::Future, hash, io, mem, pin::Pin, rc::Rc};
+use std::{cell::Cell, cell::RefCell, future::Future, hash, io, pin::Pin, rc::Rc};
 
 use slab::Slab;
 
@@ -46,11 +46,34 @@ pub(crate) struct IoStateInner {
     read_task: LocalWaker,
     write_task: LocalWaker,
     dispatch_task: LocalWaker,
-    read_buf: RefCell<BytesMut>,
-    write_buf: RefCell<BytesMut>,
+    read_buf: Cell<Option<BytesMut>>,
+    write_buf: Cell<Option<BytesMut>>,
     on_disconnect: RefCell<Slab<Option<LocalWaker>>>,
 }
 
+thread_local!(static R_BYTES_POOL: RefCell<Vec<BytesMut>> = RefCell::new(Vec::with_capacity(16)));
+thread_local!(static W_BYTES_POOL: RefCell<Vec<BytesMut>> = RefCell::new(Vec::with_capacity(16)));
+
+fn release_to_r_pool(mut buf: BytesMut) {
+    R_BYTES_POOL.with(|pool| {
+        let v = &mut pool.borrow_mut();
+        if v.len() < 16 {
+            buf.clear();
+            v.push(buf);
+        }
+    })
+}
+
+fn release_to_w_pool(mut buf: BytesMut) {
+    W_BYTES_POOL.with(|pool| {
+        let v = &mut pool.borrow_mut();
+        if v.len() < 16 {
+            buf.clear();
+            v.push(buf);
+        }
+    })
+}
+
 impl IoStateInner {
     fn insert_flags(&self, f: Flags) {
         let mut flags = self.flags.get();
@@ -63,6 +86,72 @@ impl IoStateInner {
         flags.remove(f);
         self.flags.set(flags);
     }
+
+    fn get_read_buf(&self) -> BytesMut {
+        if let Some(buf) = self.read_buf.take() {
+            buf
+        } else {
+            R_BYTES_POOL.with(|pool| {
+                if let Some(buf) = pool.borrow_mut().pop() {
+                    buf
+                } else {
+                    BytesMut::with_capacity(self.read_hw.get() as usize)
+                }
+            })
+        }
+    }
+
+    fn get_write_buf(&self) -> BytesMut {
+        if let Some(buf) = self.write_buf.take() {
+            buf
+        } else {
+            W_BYTES_POOL.with(|pool| {
+                if let Some(buf) = pool.borrow_mut().pop() {
+                    buf
+                } else {
+                    BytesMut::with_capacity(self.write_hw.get() as usize)
+                }
+            })
+        }
+    }
+
+    fn release_read_buf(&self, buf: BytesMut) {
+        if buf.is_empty() {
+            if buf.capacity() > (self.lw.get() as usize) {
+                release_to_r_pool(buf);
+            }
+        } else {
+            self.read_buf.set(Some(buf));
+        }
+    }
+
+    fn release_write_buf(&self, buf: BytesMut) {
+        if buf.is_empty() {
+            let cap = buf.capacity();
+            if cap > (self.lw.get() as usize) && cap <= self.write_hw.get() as usize {
+                release_to_w_pool(buf);
+            }
+        } else {
+            self.write_buf.set(Some(buf));
+        }
+    }
+}
+
+impl Drop for IoStateInner {
+    fn drop(&mut self) {
+        if let Some(buf) = self.read_buf.take() {
+            let cap = buf.capacity();
+            if cap > (self.lw.get() as usize) && cap <= self.read_hw.get() as usize {
+                release_to_r_pool(buf);
+            }
+        }
+        if let Some(buf) = self.write_buf.take() {
+            let cap = buf.capacity();
+            if cap > (self.lw.get() as usize) && cap <= self.write_hw.get() as usize {
+                release_to_w_pool(buf);
+            }
+        }
+    }
 }
 
 impl Clone for State {
@@ -99,8 +188,8 @@ impl State {
             dispatch_task: LocalWaker::new(),
             read_task: LocalWaker::new(),
             write_task: LocalWaker::new(),
-            read_buf: RefCell::new(BytesMut::new()),
-            write_buf: RefCell::new(BytesMut::new()),
+            read_buf: Cell::new(None),
+            write_buf: Cell::new(None),
             on_disconnect: RefCell::new(Slab::new()),
         }))
     }
@@ -109,8 +198,20 @@ impl State {
     /// Create `State` from Framed
     pub fn from_framed<Io, U>(framed: Framed<Io, U>) -> (Io, U, Self) {
         let parts = framed.into_parts();
+        let read_buf = if !parts.read_buf.is_empty() {
+            Cell::new(Some(parts.read_buf))
+        } else {
+            Cell::new(None)
+        };
+        let write_buf = if !parts.write_buf.is_empty() {
+            Cell::new(Some(parts.write_buf))
+        } else {
+            Cell::new(None)
+        };
 
         let state = State(Rc::new(IoStateInner {
+            read_buf,
+            write_buf,
             flags: Cell::new(Flags::empty()),
             error: Cell::new(None),
             lw: Cell::new(1024),
@@ -120,8 +221,6 @@ impl State {
             dispatch_task: LocalWaker::new(),
             read_task: LocalWaker::new(),
             write_task: LocalWaker::new(),
-            read_buf: RefCell::new(parts.read_buf),
-            write_buf: RefCell::new(parts.write_buf),
             on_disconnect: RefCell::new(Slab::new()),
         }));
         (parts.io, parts.codec, state)
@@ -143,10 +242,10 @@ impl State {
             write_hw: Cell::new(max_write_buf_size),
             disconnect_timeout: Cell::new(disconnect_timeout),
             dispatch_task: LocalWaker::new(),
+            read_buf: Cell::new(None),
             read_task: LocalWaker::new(),
+            write_buf: Cell::new(None),
             write_task: LocalWaker::new(),
-            read_buf: RefCell::new(BytesMut::with_capacity(min_buf_size as usize)),
-            write_buf: RefCell::new(BytesMut::with_capacity(min_buf_size as usize)),
             on_disconnect: RefCell::new(Slab::new()),
         }))
     }
@@ -155,8 +254,17 @@ impl State {
     /// Convert State to a Framed instance
     pub fn into_framed<Io, U>(self, io: Io, codec: U) -> Framed<Io, U> {
         let mut parts = FramedParts::new(io, codec);
-        parts.read_buf = mem::take(&mut self.0.read_buf.borrow_mut());
-        parts.write_buf = mem::take(&mut self.0.write_buf.borrow_mut());
+
+        parts.read_buf = if let Some(buf) = self.0.read_buf.take() {
+            buf
+        } else {
+            BytesMut::new()
+        };
+        parts.write_buf = if let Some(buf) = self.0.write_buf.take() {
+            buf
+        } else {
+            BytesMut::new()
+        };
         Framed::from_parts(parts)
     }
 
@@ -398,18 +506,15 @@ impl State {
         T: AsyncRead + AsyncWrite + Unpin,
         U: Decoder,
     {
+        let mut buf = self.0.get_read_buf();
+
         loop {
-            let item = codec.decode(&mut self.0.read_buf.borrow_mut());
-            return match item {
+            let item = codec.decode(&mut buf);
+            let result = match item {
                 Ok(Some(el)) => Ok(Some(el)),
                 Ok(None) => {
-                    let st = self.0.clone();
                     let n = poll_fn(|cx| {
-                        crate::codec::poll_read_buf(
-                            Pin::new(&mut *io),
-                            cx,
-                            &mut *st.read_buf.borrow_mut(),
-                        )
+                        crate::codec::poll_read_buf(Pin::new(&mut *io), cx, &mut buf)
                     })
                     .await
                     .map_err(Either::Right)?;
@@ -424,6 +529,8 @@ impl State {
                     Err(Either::Left(err))
                 }
             };
+            self.0.release_read_buf(buf);
+            return result;
         }
     }
 
@@ -439,10 +546,10 @@ impl State {
         T: AsyncRead + AsyncWrite + Unpin,
         U: Encoder,
     {
-        codec
-            .encode(item, &mut self.0.write_buf.borrow_mut())
-            .map_err(Either::Left)?;
+        let mut buf = self.0.get_write_buf();
+        codec.encode(item, &mut buf).map_err(Either::Left)?;
 
+        self.0.write_buf.set(Some(buf));
         if !poll_fn(|cx| self.flush_io(io, cx)).await {
             let err = self.0.error.take().unwrap_or_else(|| {
                 io::Error::new(io::ErrorKind::Other, "Internal error")
@@ -464,14 +571,13 @@ impl State {
         T: AsyncRead + AsyncWrite + Unpin,
         U: Decoder,
     {
-        let mut buf = self.0.read_buf.borrow_mut();
+        let mut buf = self.0.get_read_buf();
 
         loop {
-            return match codec.decode(&mut buf) {
+            let item = match codec.decode(&mut buf) {
                 Ok(Some(el)) => Poll::Ready(Ok(Some(el))),
                 Ok(None) => {
-                    match crate::codec::poll_read_buf(Pin::new(&mut *io), cx, &mut *buf)
-                    {
+                    match crate::codec::poll_read_buf(Pin::new(&mut *io), cx, &mut buf) {
                         Poll::Pending => Poll::Pending,
                         Poll::Ready(Err(err)) => Poll::Ready(Err(Either::Right(err))),
                         Poll::Ready(Ok(n)) => {
@@ -488,6 +594,8 @@ impl State {
                     Poll::Ready(Err(Either::Left(err)))
                 }
             };
+            self.0.release_read_buf(buf);
+            return item;
         }
     }
 
@@ -498,7 +606,7 @@ impl State {
     {
         let inner = self.0.as_ref();
         let lw = inner.lw.get() as usize;
-        let buf = &mut inner.read_buf.borrow_mut();
+        let mut buf = inner.get_read_buf();
 
         // read data from socket
         let mut updated = false;
@@ -509,11 +617,12 @@ impl State {
                 buf.reserve((inner.read_hw.get() as usize) - remaining);
             }
 
-            match crate::codec::poll_read_buf(Pin::new(&mut *io), cx, buf) {
+            match crate::codec::poll_read_buf(Pin::new(&mut *io), cx, &mut buf) {
                 Poll::Pending => break,
                 Poll::Ready(Ok(n)) => {
                     if n == 0 {
                         log::trace!("io stream is disconnected");
+                        inner.release_read_buf(buf);
                         self.set_io_error(None);
                         return false;
                     } else {
@@ -522,9 +631,10 @@ impl State {
                                 "buffer is too large {}, enable read back-pressure",
                                 buf.len()
                             );
-                            self.insert_flags(Flags::RD_READY | Flags::RD_BUF_FULL);
-                            self.0.dispatch_task.wake();
-                            self.0.read_task.register(cx.waker());
+                            inner.dispatch_task.wake();
+                            inner.read_buf.set(Some(buf));
+                            inner.read_task.register(cx.waker());
+                            inner.insert_flags(Flags::RD_READY | Flags::RD_BUF_FULL);
                             return true;
                         }
 
@@ -533,6 +643,7 @@ impl State {
                 }
                 Poll::Ready(Err(err)) => {
                     log::trace!("read task failed on io {:?}", err);
+                    inner.release_read_buf(buf);
                     self.set_io_error(Some(err));
                     return false;
                 }
@@ -540,8 +651,11 @@ impl State {
         }
 
         if updated {
+            inner.read_buf.set(Some(buf));
             self.insert_flags(Flags::RD_READY);
             self.0.dispatch_task.wake();
+        } else {
+            inner.release_read_buf(buf);
         }
         self.0.read_task.register(cx.waker());
         true
@@ -553,7 +667,12 @@ impl State {
         T: AsyncRead + AsyncWrite + Unpin,
     {
         let inner = self.0.as_ref();
-        let buf = &mut inner.write_buf.borrow_mut();
+        let mut buf = if let Some(buf) = inner.write_buf.take() {
+            buf
+        } else {
+            self.0.write_task.register(cx.waker());
+            return Poll::Ready(true);
+        };
         let len = buf.len();
 
         if len != 0 {
@@ -569,6 +688,8 @@ impl State {
                                 "Disconnected during flush, written {}",
                                 written
                             );
+                            buf.clear();
+                            inner.release_write_buf(buf);
                             self.set_io_error(Some(io::Error::new(
                                 io::ErrorKind::WriteZero,
                                 "failed to write frame to transport",
@@ -580,6 +701,8 @@ impl State {
                     }
                     Poll::Ready(Err(e)) => {
                         log::trace!("Error during flush: {}", e);
+                        buf.clear();
+                        inner.release_write_buf(buf);
                         self.set_io_error(Some(e));
                         return Poll::Ready(false);
                     }
@@ -609,7 +732,7 @@ impl State {
         self.0.write_task.register(cx.waker());
 
         // flush
-        match Pin::new(&mut *io).poll_flush(cx) {
+        let result = match Pin::new(&mut *io).poll_flush(cx) {
             Poll::Ready(Ok(_)) => {
                 if buf.is_empty() {
                     Poll::Ready(true)
@@ -623,7 +746,9 @@ impl State {
                 self.set_io_error(Some(err));
                 Poll::Ready(false)
             }
-        }
+        };
+        inner.release_write_buf(buf);
+        result
     }
 }
 
@@ -640,7 +765,13 @@ impl<'a> Write<'a> {
     #[inline]
     /// Check if write buffer is full
     pub fn is_full(&self) -> bool {
-        self.0.write_buf.borrow().len() >= self.0.write_hw.get() as usize
+        if let Some(buf) = self.0.read_buf.take() {
+            let result = buf.len() >= self.0.write_hw.get() as usize;
+            self.0.write_buf.set(Some(buf));
+            result
+        } else {
+            false
+        }
     }
 
     #[inline]
@@ -666,12 +797,14 @@ impl<'a> Write<'a> {
     where
         F: FnOnce(&mut BytesMut) -> R,
     {
-        let mut write_buf = self.0.write_buf.borrow_mut();
-        if write_buf.is_empty() {
+        let mut buf = self.0.get_write_buf();
+        if buf.is_empty() {
             self.0.write_task.wake();
         }
 
-        f(&mut write_buf)
+        let result = f(&mut buf);
+        self.0.release_write_buf(buf);
+        result
     }
 
     #[inline]
@@ -689,22 +822,24 @@ impl<'a> Write<'a> {
         let flags = self.0.flags.get();
 
         if !flags.intersects(Flags::IO_ERR | Flags::IO_SHUTDOWN) {
-            let mut write_buf = self.0.write_buf.borrow_mut();
-            let is_write_sleep = write_buf.is_empty();
+            let mut buf = self.0.get_write_buf();
+            let is_write_sleep = buf.is_empty();
 
             // make sure we've got room
-            let remaining = write_buf.capacity() - write_buf.len();
+            let remaining = buf.capacity() - buf.len();
             if remaining < self.0.lw.get() as usize {
-                write_buf.reserve((self.0.write_hw.get() as usize) - remaining);
+                buf.reserve((self.0.write_hw.get() as usize) - remaining);
             }
 
             // encode item and wake write task
-            codec.encode(item, &mut *write_buf).map(|_| {
+            let result = codec.encode(item, &mut buf).map(|_| {
                 if is_write_sleep {
                     self.0.write_task.wake();
                 }
-                write_buf.len() < self.0.write_hw.get() as usize
-            })
+                buf.len() < self.0.write_hw.get() as usize
+            });
+            self.0.write_buf.set(Some(buf));
+            result
         } else {
             Ok(true)
         }
@@ -725,25 +860,28 @@ impl<'a> Write<'a> {
         if !flags.intersects(Flags::IO_ERR | Flags::ST_DSP_ERR) {
             match item {
                 Ok(Some(item)) => {
-                    let mut write_buf = self.0.write_buf.borrow_mut();
-                    let is_write_sleep = write_buf.is_empty();
+                    let mut buf = self.0.get_write_buf();
+                    let is_write_sleep = buf.is_empty();
 
                     // make sure we've got room
-                    let remaining = write_buf.capacity() - write_buf.len();
+                    let remaining = buf.capacity() - buf.len();
                     if remaining < self.0.lw.get() as usize {
-                        write_buf.reserve((self.0.write_hw.get() as usize) - remaining);
+                        buf.reserve((self.0.write_hw.get() as usize) - remaining);
                     }
 
                     // encode item
-                    if let Err(err) = codec.encode(item, &mut write_buf) {
+                    if let Err(err) = codec.encode(item, &mut buf) {
                         log::trace!("Encoder error: {:?}", err);
+                        self.0.release_write_buf(buf);
                         self.0.insert_flags(Flags::DSP_STOP | Flags::ST_DSP_ERR);
                         self.0.dispatch_task.wake();
                         return Err(Either::Right(err));
                     } else if is_write_sleep {
                         self.0.write_task.wake();
                     }
-                    Ok(write_buf.len() < self.0.write_hw.get() as usize)
+                    let result = Ok(buf.len() < self.0.write_hw.get() as usize);
+                    self.0.write_buf.set(Some(buf));
+                    result
                 }
                 Err(err) => {
                     self.0.insert_flags(Flags::DSP_STOP | Flags::ST_DSP_ERR);
@@ -771,7 +909,13 @@ impl<'a> Read<'a> {
     #[inline]
     /// Check if read buffer is full
     pub fn is_full(&self) -> bool {
-        self.0.read_buf.borrow().len() >= self.0.read_hw.get() as usize
+        if let Some(buf) = self.0.read_buf.take() {
+            let result = buf.len() >= self.0.read_hw.get() as usize;
+            self.0.read_buf.set(Some(buf));
+            result
+        } else {
+            false
+        }
     }
 
     #[inline]
@@ -819,7 +963,13 @@ impl<'a> Read<'a> {
     where
         U: Decoder,
     {
-        codec.decode(&mut self.0.read_buf.borrow_mut())
+        if let Some(mut buf) = self.0.read_buf.take() {
+            let result = codec.decode(&mut buf);
+            self.0.release_read_buf(buf);
+            result
+        } else {
+            codec.decode(&mut BytesMut::new())
+        }
     }
 
     /// Get mut access to read buffer
@@ -827,7 +977,13 @@ impl<'a> Read<'a> {
     where
         F: FnOnce(&mut BytesMut) -> R,
     {
-        f(&mut self.0.read_buf.borrow_mut())
+        if let Some(mut buf) = self.0.read_buf.take() {
+            let res = f(&mut buf);
+            self.0.release_read_buf(buf);
+            res
+        } else {
+            f(&mut BytesMut::new())
+        }
     }
 }
 
