@@ -15,7 +15,7 @@ use super::service::{Factory, InternalServiceFactory, StreamServiceFactory};
 use super::signals::{Signal, Signals};
 use super::socket::Listener;
 use super::worker::{self, Worker, WorkerAvailability, WorkerClient};
-use super::{Server, ServerCommand, Token};
+use super::{Server, ServerCommand, ServerStatus, Token};
 
 const STOP_DELAY: Duration = Duration::from_millis(300);
 
@@ -128,6 +128,18 @@ impl ServerBuilder {
         self
     }
 
+    #[doc(hidden)]
+    /// Set server status handler.
+    ///
+    /// Server calls this handler on every inner status update.
+    pub fn status_handler<F>(mut self, handler: F) -> Self
+    where
+        F: FnMut(ServerStatus) + Send + 'static,
+    {
+        self.accept.set_status_handler(handler);
+        self
+    }
+
     /// Execute external configuration as part of the server building
     /// process.
     ///
